@@ -18,12 +18,20 @@ from functools import reduce
 import itertools
 from lxml import etree  # pytype: disable=import-error
 import re
-from typing import List, Optional, Tuple
-from picosvg.svg_meta import ntos, svgns, xlinkns, parse_css_declarations
+from typing import List, Optional, Sequence, Tuple
+from picosvg.svg_meta import (
+    number_or_percentage,
+    ntos,
+    strip_ns,
+    svgns,
+    xlinkns,
+    parse_css_declarations,
+)
 from picosvg.svg_types import *
+from picosvg.svg_transform import Affine2D
 import numbers
 
-_ELEMENT_CLASSES = {
+_SHAPE_CLASSES = {
     "circle": SVGCircle,
     "ellipse": SVGEllipse,
     "line": SVGLine,
@@ -32,8 +40,17 @@ _ELEMENT_CLASSES = {
     "polyline": SVGPolyline,
     "rect": SVGRect,
 }
-_CLASS_ELEMENTS = {v: f"{{{svgns()}}}{k}" for k, v in _ELEMENT_CLASSES.items()}
-_ELEMENT_CLASSES.update({f"{{{svgns()}}}{k}": v for k, v in _ELEMENT_CLASSES.items()})
+_CLASS_ELEMENTS = {v: f"{{{svgns()}}}{k}" for k, v in _SHAPE_CLASSES.items()}
+_SHAPE_CLASSES.update({f"{{{svgns()}}}{k}": v for k, v in _SHAPE_CLASSES.items()})
+
+_GRADIENT_CLASSES = {
+    "linearGradient": SVGLinearGradient,
+    "radialGradient": SVGRadialGradient,
+}
+_GRADIENT_COORDS = {
+    "linearGradient": (("x1", "y1"), ("x2", "y2")),
+    "radialGradient": (("cx", "cy"), ("fx", "fy")),
+}
 
 _XLINK_TEMP = "xlink_"
 
@@ -46,6 +63,10 @@ _MAX_PCT_ERROR = 0.1
 _DEFAULT_DEFAULT_TOLERENCE = 0.1
 
 
+# Rounding for rewritten gradient matrices
+_GRADIENT_TRANSFORM_NDIGITS = 6
+
+
 def _xlink_href_attr_name() -> str:
     return f"{{{xlinkns()}}}href"
 
@@ -104,9 +125,10 @@ def _field_name(attr_name):
 
 
 def from_element(el):
-    if el.tag not in _ELEMENT_CLASSES:
+    if el.tag not in _SHAPE_CLASSES:
         raise ValueError(f"Bad tag <{el.tag}>")
-    data_type = _ELEMENT_CLASSES[el.tag]
+    data_type = _SHAPE_CLASSES[el.tag]
+    parse_fn = getattr(data_type, "from_element", None)
     args = {
         f.name: f.type(el.attrib[_attr_name(f.name)])
         for f in dataclasses.fields(data_type)
@@ -135,6 +157,12 @@ def _reset_attrs(data_obj, field_pred):
             setattr(data_obj, field.name, field.default)
 
 
+def _safe_remove(el: etree.Element):
+    parent = el.getparent()
+    if parent is not None:
+        parent.remove(el)
+
+
 class SVG:
 
     svg_root: etree.Element
@@ -148,8 +176,9 @@ class SVG:
         if self.elements:
             return self.elements
         elements = []
+        view_box = self.view_box()
         for el in self.svg_root.iter("*"):
-            if el.tag not in _ELEMENT_CLASSES:
+            if el.tag not in _SHAPE_CLASSES:
                 continue
             elements.append((el, (from_element(el),)))
         self.elements = elements
@@ -293,27 +322,27 @@ class SVG:
         self._resolve_use(self.svg_root)
         return self
 
-    def _resolve_clip_path(self, clip_path_url):
+    def _resolve_clip_path(self, clip_path_url) -> SVGPath:
         clip_path_el = self.resolve_url(clip_path_url, "clipPath")
         self._resolve_use(clip_path_el)
         self._ungroup(clip_path_el)
 
         # union all the shapes under the clipPath
         # Fails if there are any non-shapes under clipPath
-        return union([from_element(e) for e in clip_path_el])
+        return SVGPath.from_commands(union([from_element(e) for e in clip_path_el]))
 
     def append_to(self, xpath, el):
         self._update_etree()
         self.xpath_one(xpath).append(el)
         return el
 
-    def _combine_clip_paths(self, clip_paths) -> SVGPath:
+    def _combine_clip_paths(self, clip_paths: Sequence[SVGPath]) -> SVGPath:
         # multiple clip paths leave behind their intersection
         if not clip_paths:
             raise ValueError("Cannot combine no clip_paths")
         if len(clip_paths) == 1:
             return clip_paths[0]
-        return intersection(clip_paths)
+        return SVGPath.from_commands(intersection(clip_paths))
 
     def _new_id(self, tag, template):
         for i in range(100):
@@ -343,11 +372,22 @@ class SVG:
             if value != "visible":
                 _inherit_copy(attrib, child, attr_name)
 
+        def _inherit_matrix_multiply(attrib, child, attr_name):
+            group_transform = Affine2D.fromstring(attrib[attr_name])
+            if attr_name in child.attrib:
+                transform = Affine2D.fromstring(child.attrib[attr_name])
+                transform = Affine2D.product(transform, group_transform)
+            else:
+                transform = group_transform
+            if transform != Affine2D.identity():
+                child.attrib[attr_name] = transform.tostring()
+
         attrib_handlers = {
             "clip-rule": _inherit_copy,
             "fill": _inherit_copy,
             "fill-rule": _inherit_copy,
             "style": _inherit_copy,
+            "transform": _inherit_matrix_multiply,
             "stroke": _inherit_copy,
             "stroke-width": _inherit_copy,
             "stroke-linecap": _inherit_copy,
@@ -383,8 +423,7 @@ class SVG:
         # nuke the groups that are not displayed
         display_none = [e for e in self.xpath(f".//svg:g[@display='none']", scope_el)]
         for group in display_none:
-            if group.getparent() is not None:
-                group.getparent().remove(group)
+            _safe_remove(group)
 
         # Any groups left are displayed
         groups = [e for e in self.xpath(f".//svg:g", scope_el)]
@@ -407,8 +446,7 @@ class SVG:
 
         # nuke the groups
         for group in groups:
-            if group.getparent() is not None:
-                group.getparent().remove(group)
+            _safe_remove(group)
 
         # if we have new combinations of clip paths dedup & materialize them
         new_clip_paths = {}
@@ -438,7 +476,7 @@ class SVG:
                 continue
             old_id = old_clip_path.attrib["id"]
             if not self.xpath(f'//svg:*[@clip-path="url(#{old_id})"]'):
-                old_clip_path.getparent().remove(old_clip_path)
+                _safe_remove(old_clip_path)
 
     def _compute_clip_path(self, el):
         """Resolve clip path for element, including inherited clipping.
@@ -474,8 +512,7 @@ class SVG:
         Returns sequence of shapes in draw order. That is, result[1] should be
         drawn on top of result[0], etc."""
 
-        if shape.stroke == "none":
-            return (shape,)
+        assert shape.stroke != "none"
 
         # make a new path that is the stroke
         stroke = shape.as_path().update_path(shape.stroke_commands(self.tolerance))
@@ -486,6 +523,10 @@ class SVG:
         # a few attributes move in interesting ways
         stroke.opacity *= stroke.stroke_opacity
         stroke.fill = stroke.stroke
+        # the fill and stroke are now different (filled) paths, reset 'fill_opacity'
+        # to default and only use a combined 'opacity' in each one.
+        shape.opacity *= shape.fill_opacity
+        shape.fill_opacity = stroke.fill_opacity = 1.0
 
         # remove all the stroke settings
         for cleanmeup in (shape, stroke):
@@ -555,13 +596,55 @@ class SVG:
 
         # destroy clip path elements
         for clip_path_el in self.xpath("//svg:clipPath"):
-            clip_path_el.getparent().remove(clip_path_el)
+            _safe_remove(clip_path_el)
 
         # destroy clip-path attributes
         self.remove_attributes(["clip-path"], xpath="//svg:*[@clip-path]", inplace=True)
 
         return self
 
+    def clip_to_viewbox(self, inplace=False):
+        if not inplace:
+            svg = SVG(copy.deepcopy(self.svg_root))
+            svg.clip_to_viewbox(inplace=True)
+            return svg
+
+        self._update_etree()
+
+        view_box = self.view_box()
+
+        # phase 1: dump shapes that are completely out of bounds
+        for el, (shape,) in self._elements():
+            if view_box.intersection(shape.bounding_box()) is None:
+                _safe_remove(el)
+
+        self.elements = None  # force elements to reload
+
+        # phase 2: clip things that are partially out of bounds
+        updates = []
+        for idx, (el, (shape,)) in enumerate(self._elements()):
+            bbox = shape.bounding_box()
+            isct = view_box.intersection(bbox)
+            assert isct is not None, f"We should have already dumped {shape}"
+            if bbox == isct:
+                continue
+            clip_path = (
+                SVGRect(x=isct.x, y=isct.y, width=isct.w, height=isct.h)
+                .as_path()
+                .absolute(inplace=True)
+            )
+            shape = shape.as_path().absolute(inplace=True)
+            shape.update_path(intersection((shape, clip_path)), inplace=True)
+            updates.append((idx, el, shape))
+
+        for idx, el, shape in updates:
+            self._set_element(idx, el, (shape,))
+
+        # Update the etree
+        self._update_etree()
+
+        return self
+
     def apply_transforms(self, inplace=False):
         """Naively transforms to shapes and removes the transform attribute.
 
@@ -651,6 +734,21 @@ class SVG:
 
         return self
 
+    def remove_anonymous_symbols(self, inplace=False):
+        # No id makes a symbol useless
+        # https://github.com/googlefonts/picosvg/issues/46
+        if not inplace:
+            svg = SVG(copy.deepcopy(self.svg_root))
+            svg.remove_anonymous_symbols(inplace=True)
+            return svg
+
+        self._update_etree()
+
+        for el in self.xpath("//svg:symbol[not(@id)]"):
+            el.getparent().remove(el)
+
+        return self
+
     def set_attributes(self, name_values, xpath="/svg:svg", inplace=False):
         if not inplace:
             svg = SVG(copy.deepcopy(self.svg_root))
@@ -679,17 +777,117 @@ class SVG:
 
         return self
 
+    def normalize_opacity(self, inplace=False):
+        """Merge '{fill,stroke}_opacity' with generic 'opacity' when possible."""
+        if not inplace:
+            svg = SVG(copy.deepcopy(self.svg_root))
+            svg.normalize_opacity(inplace=True)
+            return svg
+
+        for shape in self.shapes():
+            shape.normalize_opacity(inplace=True)
+
+        return self
+
+    def _select_gradients(self):
+        return self.xpath(" | ".join(f"//svg:{tag}" for tag in _GRADIENT_CLASSES))
+
+    def _collect_gradients(self, inplace=False):
+        if not inplace:
+            svg = SVG(copy.deepcopy(self.svg_root))
+            svg._collect_gradients(inplace=False)
+            return svg
+
+        # Collect gradients; remove other defs
+        defs = etree.Element(f"{{{svgns()}}}defs", nsmap=self.svg_root.nsmap)
+        for gradient in self._select_gradients():
+            gradient.getparent().remove(gradient)
+            defs.append(gradient)
+
+        for def_el in [e for e in self.xpath("//svg:defs")]:
+            def_el.getparent().remove(def_el)
+
+        self.svg_root.insert(0, defs)
+
+    def _apply_gradient_translation(self, inplace=False):
+        if not inplace:
+            svg = SVG(copy.deepcopy(self.svg_root))
+            svg._apply_gradient_translation(inplace=True)
+            return svg
+
+        for el in self._select_gradients():
+            gradient = _GRADIENT_CLASSES[strip_ns(el.tag)].from_element(
+                el, self.view_box()
+            )
+            a, b, c, d, e, f = (
+                round(v, _GRADIENT_TRANSFORM_NDIGITS)
+                for v in gradient.gradientTransform
+            )
+            affine = Affine2D(a, b, c, d, e, f)
+            #  no translate? nop!
+            if (e, f) == (0, 0):
+                continue
+            affine_prime = affine._replace(e=0, f=0)
+
+            for x_attr, y_attr in _GRADIENT_COORDS[strip_ns(el.tag)]:
+                # if at default just ignore
+                if x_attr not in el.attrib and y_attr not in el.attrib:
+                    continue
+                x = getattr(gradient, x_attr)
+                y = getattr(gradient, y_attr)
+
+                # We need x`, y` such that matrix a b c d 0 0 yields same
+                # result as x,y with a b c d e f
+                # That is:
+                # 1)  ax` + cy` + 0 = ax + cy + e
+                # 2)  bx` + dy` + 0 = bx + dy + f
+                #                   ^ rhs is a known scalar; we'll call r1, r2
+                # multiply 1) by b/a so when subtracted from 2) we eliminate x`
+                # 1)  bx` + (b/a)cy` = (b/a) * r1
+                # 2) - 1)  bx` - bx` + dy` - (b/a)cy` = r2 - (b/a) * r1
+                #         y` = (r2 - (b/a) * r1) / (d - (b/a)c)
+                r1, r2 = affine.map_point((x, y))
+                assert r1 == a * x + c * y + e
+                assert r2 == b * x + d * y + f
+
+                if a != 0:
+                    y_prime = (r2 - r1 * b / a) / (d - b * c / a)
+
+                    # Sub y` into 1)
+                    # 1) x` = (r1 - cy`) / a
+                    x_prime = (r1 - c * y_prime) / a
+                else:
+                    # if a == 0 then above gives div / 0. Take a simpler path.
+                    # 1) 0x` + cy` + 0 = 0x + cy + e
+                    #    y` = y + e/c
+                    y_prime = y + e / c
+                    # Sub y` into 2)
+                    # 2)  bx` + dy` + 0 = bx + dy + f
+                    #      x` = x + dy/b  + f/b - dy`/b
+                    x_prime = x + (d * y / b) + (f / b) - (d * y_prime / b)
+
+                # sanity check: a`(x`, y`) should be a(x, y)
+                # all our float brutality damages points; low tolerance sanity checks!
+                p = Point(r1, r2)
+                p_prime = affine_prime.map_point((x_prime, y_prime))
+                assert p.almost_equals(p_prime, tolerance=1e-1), f"{p} != {p_prime}"
+
+                el.attrib[x_attr] = ntos(round(x_prime, _GRADIENT_TRANSFORM_NDIGITS))
+                el.attrib[y_attr] = ntos(round(y_prime, _GRADIENT_TRANSFORM_NDIGITS))
+
+            if affine_prime != Affine2D.identity():
+                el.attrib["gradientTransform"] = (
+                    "matrix(" + " ".join(ntos(v) for v in affine_prime) + ")"
+                )
+            else:
+                del el.attrib["gradientTransform"]
+
     def checkpicosvg(self):
         """Check for nano violations, return xpaths to bad elements.
 
         If result sequence empty then this is a valid picosvg.
         """
 
-        def _strip_ns(tagname):
-            if "}" in tagname:
-                return tagname[tagname.index("}") + 1 :]
-            return tagname
-
         self._update_etree()
 
         errors = []
@@ -705,11 +903,12 @@ class SVG:
         frontier = [(0, self.svg_root, "")]
         while frontier:
             el_idx, el, parent_path = frontier.pop(0)
-            el_tag = _strip_ns(el.tag)
+            el_tag = strip_ns(el.tag)
             el_path = f"{parent_path}/{el_tag}[{el_idx}]"
 
             if not any((re.match(pat, el_path) for pat in path_whitelist)):
                 errors.append(f"BadElement: {el_path}")
+                continue  # no sense reporting all the children as bad
 
             for child_idx, child in enumerate(el):
                 if child.tag is etree.Comment:
@@ -729,29 +928,23 @@ class SVG:
         self._update_etree()
 
         self.remove_comments(inplace=True)
+        self.remove_anonymous_symbols(inplace=True)
         self.apply_style_attributes(inplace=True)
         self.shapes_to_paths(inplace=True)
         self.resolve_use(inplace=True)
-        self.apply_transforms(inplace=True)
         self.ungroup(inplace=True)
         # stroke after ungroup to apply group strokes properly
         self.strokes_to_paths(inplace=True)
+        self.apply_transforms(inplace=True)
         self.apply_clip_paths(inplace=True)
         self.evenodd_to_nonzero_winding(inplace=True)
         self.remove_unpainted_shapes(inplace=True)
+        self.normalize_opacity(inplace=True)
         self.absolute(inplace=True)
         self.round_floats(ndigits, inplace=True)
 
-        # Collect gradients; remove other defs
-        defs = etree.Element(f"{{{svgns()}}}defs", nsmap=self.svg_root.nsmap)
-        for gradient in self.xpath("//svg:linearGradient | //svg:radialGradient"):
-            gradient.getparent().remove(gradient)
-            defs.append(gradient)
-
-        for def_el in [e for e in self.xpath("//svg:defs")]:
-            def_el.getparent().remove(def_el)
-
-        self.svg_root.insert(0, defs)
+        self._apply_gradient_translation(inplace=True)
+        self._collect_gradients(inplace=True)
 
         nano_violations = self.checkpicosvg()
         if nano_violations:
