@@ -17,7 +17,7 @@ from lxml import etree
 import os
 import pytest
 from picosvg.svg import SVG
-from picosvg import svg_meta
+from picosvg.svg_meta import strip_ns, parse_css_declarations
 from svg_test_helpers import *
 
 
@@ -75,48 +75,48 @@ def test_common_attrib(shape, expected_fields):
             'd="M10,110.2 L50.5,150.7"',
         ),
         # rect: minimal valid example
-        ("<rect width='1' height='1'/>", 'd="M0,0 H1 V1 H0 V0 z"'),
+        ("<rect width='1' height='1'/>", 'd="M0,0 H1 V1 H0 V0 Z"'),
         # rect: sharp corners
         (
             "<rect x='10' y='11' width='17' height='11'/>",
-            'd="M10,11 H27 V22 H10 V11 z"',
+            'd="M10,11 H27 V22 H10 V11 Z"',
         ),
         # rect: round corners
         (
             "<rect x='9' y='9' width='11' height='7' rx='2'/>",
             'd="M11,9 H18 A2 2 0 0 1 20,11 V14 A2 2 0 0 1 18,16 H11'
-            ' A2 2 0 0 1 9,14 V11 A2 2 0 0 1 11,9 z"',
+            ' A2 2 0 0 1 9,14 V11 A2 2 0 0 1 11,9 Z"',
         ),
         # rect: simple
         (
             "<rect x='11.5' y='16' width='11' height='2'/>",
-            'd="M11.5,16 H22.5 V18 H11.5 V16 z"',
+            'd="M11.5,16 H22.5 V18 H11.5 V16 Z"',
         ),
         # polygon
-        ("<polygon points='30,10 50,30 10,30'/>", 'd="M30,10 50,30 10,30 z"'),
+        ("<polygon points='30,10 50,30 10,30'/>", 'd="M30,10 50,30 10,30 Z"'),
         # polyline
         ("<polyline points='30,10 50,30 10,30'/>", 'd="M30,10 50,30 10,30"'),
         # circle, minimal valid example
-        ("<circle r='1'/>", 'd="M-1,0 A1 1 0 1 1 1,0 A1 1 0 1 1 -1,0 z"'),
+        ("<circle r='1'/>", 'd="M-1,0 A1 1 0 1 1 1,0 A1 1 0 1 1 -1,0 Z"'),
         # circle
         (
             "<circle cx='600' cy='200' r='100'/>",
-            'd="M500,200 A100 100 0 1 1 700,200 A100 100 0 1 1 500,200 z"',
+            'd="M500,200 A100 100 0 1 1 700,200 A100 100 0 1 1 500,200 Z"',
         ),
         # circle, decimal positioning
         (
             "<circle cx='12' cy='6.5' r='1.5'></circle>",
-            'd="M10.5,6.5 A1.5 1.5 0 1 1 13.5,6.5 A1.5 1.5 0 1 1 10.5,6.5 z"',
+            'd="M10.5,6.5 A1.5 1.5 0 1 1 13.5,6.5 A1.5 1.5 0 1 1 10.5,6.5 Z"',
         ),
         # ellipse
         (
             '<ellipse cx="100" cy="50" rx="100" ry="50"/>',
-            'd="M0,50 A100 50 0 1 1 200,50 A100 50 0 1 1 0,50 z"',
+            'd="M0,50 A100 50 0 1 1 200,50 A100 50 0 1 1 0,50 Z"',
         ),
         # ellipse, decimal positioning
         (
             '<ellipse cx="100.5" cy="50" rx="10" ry="50.5"/>',
-            'd="M90.5,50 A10 50.5 0 1 1 110.5,50 A10 50.5 0 1 1 90.5,50 z"',
+            'd="M90.5,50 A10 50.5 0 1 1 110.5,50 A10 50.5 0 1 1 90.5,50 Z"',
         ),
     ],
 )
@@ -222,6 +222,8 @@ def test_ungroup(actual, expected_result):
         ("stroke-simplepath-before.svg", "stroke-simplepath-after.svg"),
         ("stroke-path-before.svg", "stroke-path-after.svg"),
         ("stroke-capjoinmiterlimit-before.svg", "stroke-capjoinmiterlimit-after.svg"),
+        ("scale-strokes-before.svg", "scale-strokes-after.svg"),
+        ("stroke-fill-opacity-before.svg", "stroke-fill-opacity-after.svg"),
     ],
 )
 def test_strokes_to_paths(actual, expected_result):
@@ -263,12 +265,24 @@ def test_transform(actual, expected_result):
         ("twemoji-lesotho-flag-before.svg", "twemoji-lesotho-flag-nano.svg"),
         ("inline-css-style-before.svg", "inline-css-style-nano.svg"),
         ("clipped-strokes-before.svg", "clipped-strokes-nano.svg"),
+        ("drop-anon-symbols-before.svg", "drop-anon-symbols-after.svg"),
+        ("scale-strokes-before.svg", "scale-strokes-nano.svg"),
     ],
 )
 def test_topicosvg(actual, expected_result):
     _test(actual, expected_result, lambda svg: svg.topicosvg())
 
 
+@pytest.mark.parametrize(
+    "actual, expected_result",
+    [
+        ("outside-viewbox.svg", "outside-viewbox-clipped.svg"),
+    ],
+)
+def test_clip_to_viewbox(actual, expected_result):
+    _test(actual, expected_result, lambda svg: svg.clip_to_viewbox().round_floats(4))
+
+
 @pytest.mark.parametrize(
     "actual, expected_result", [("invisible-before.svg", "invisible-after.svg")]
 )
@@ -285,7 +299,6 @@ def test_remove_unpainted_shapes(actual, expected_result):
             (
                 "BadElement: /svg[0]/defs[1]",
                 "BadElement: /svg[0]/donkey[2]",
-                "BadElement: /svg[0]/defs[1]/path[0]",
             ),
         ),
         ("bad-defs-1.svg", ("BadElement: /svg[0]/path[0]",)),
@@ -380,7 +393,7 @@ def test_parse_css_declarations(
     style, property_names, expected_output, expected_unparsed
 ):
     output = {}
-    unparsed = svg_meta.parse_css_declarations(style, output, property_names)
+    unparsed = parse_css_declarations(style, output, property_names)
     assert output == expected_output
     assert unparsed == expected_unparsed
 
@@ -388,7 +401,7 @@ def test_parse_css_declarations(
 @pytest.mark.parametrize("style", ["foo;bar;", "foo:bar:baz;"])
 def test_parse_css_declarations_invalid(style):
     with pytest.raises(ValueError, match="Invalid CSS declaration syntax"):
-        svg_meta.parse_css_declarations(style, {})
+        parse_css_declarations(style, {})
 
 
 @pytest.mark.parametrize(
@@ -403,3 +416,64 @@ def test_apply_style_attributes(actual, expected_result):
         expected_result,
         lambda svg: svg.shapes() and svg.apply_style_attributes(),
     )
+
+
+@pytest.mark.parametrize(
+    "gradient_string, expected_result",
+    [
+        # No transform, no change
+        (
+            '<linearGradient id="c" x1="63.85" x2="63.85" y1="4245" y2="4137.3" gradientUnits="userSpaceOnUse"/>',
+            '<linearGradient id="c" x1="63.85" x2="63.85" y1="4245" y2="4137.3" gradientUnits="userSpaceOnUse"/>',
+        ),
+        # Real example from emoji_u1f392.svg w/ dx changed from 0 to 1
+        # scale, translate
+        (
+            '<linearGradient id="c" x1="63.85" x2="63.85" y1="4245" y2="4137.3" gradientTransform="translate(1 -4122)" gradientUnits="userSpaceOnUse"/>',
+            '<linearGradient id="c" x1="64.85" x2="64.85" y1="123" y2="15.3" gradientUnits="userSpaceOnUse"/>',
+        ),
+        # Real example from emoji_u1f392.svg w/sx changed from 1 to 0.5
+        # scale, translate
+        (
+            '<radialGradient id="b" cx="63.523" cy="12368" r="53.477" gradientTransform="matrix(.5 0 0 .2631 0 -3150)" gradientUnits="userSpaceOnUse"/>',
+            '<radialGradient id="b" cx="63.523" cy="395.366021" r="53.477" gradientTransform="matrix(0.5 0 0 0.2631 0 0)" gradientUnits="userSpaceOnUse"/>',
+        ),
+        # Real example from emoji_u1f44d.svg
+        # Using all 6 parts
+        (
+            '<radialGradient id="d" cx="2459.4" cy="-319.18" r="20.331" gradientTransform="matrix(-1.3883 .0794 -.0374 -.6794 3505.4 -353.39)" gradientUnits="userSpaceOnUse"/>',
+            '<radialGradient id="d" cx="-71.60264" cy="-94.82264" r="20.331" gradientTransform="matrix(-1.3883 0.0794 -0.0374 -0.6794 0 0)" gradientUnits="userSpaceOnUse"/>',
+        ),
+        # Manually constructed objectBBox
+        (
+            '<radialGradient id="mbbox" cx="0.75" cy="0.75" r="0.40" gradientTransform="matrix(1 1 -0.7873 -0.001717 0.5 0)" gradientUnits="objectBoundingBox"/>',
+            '<radialGradient id="mbbox" cx="0.748907" cy="0.11353" r="0.40" gradientTransform="matrix(1 1 -0.7873 -0.001717 0 0)" gradientUnits="objectBoundingBox"/>',
+        ),
+        # Real example from emoji_u26BE
+        # https://github.com/googlefonts/picosvg/issues/129
+        (
+            '<radialGradient id="f" cx="-779.79" cy="3150" r="58.471" gradientTransform="matrix(0 1 -1 0 3082.5 1129.5)" gradientUnits="userSpaceOnUse"/>',
+            '<radialGradient id="f" cx="349.71" cy="67.5" r="58.471" gradientTransform="matrix(0 1 -1 0 0 0)" gradientUnits="userSpaceOnUse"/>',
+        ),
+        # Real example from emoji_u270c.svg
+        # Very small values (e-17...) and float math makes for large errors
+        (
+            '<radialGradient id="f" cx="75.915" cy="20.049" r="71.484" fx="88.617" fy="-50.297" gradientTransform="matrix(6.1232e-17 1 -1.0519 6.4408e-17 97.004 -55.866)" gradientUnits="userSpaceOnUse"/>',
+            '<radialGradient id="f" cx="20.049" cy="-72.168891" r="71.484" fx="32.751" fy="-142.514891" gradientTransform="matrix(0 1 -1.0519 0 0 0)" gradientUnits="userSpaceOnUse"/>',
+        ),
+    ],
+)
+def test_apply_gradient_translation(gradient_string, expected_result):
+    svg_string = (
+        '<svg version="1.1" xmlns="http://www.w3.org/2000/svg" viewBox="0 0 128 128">'
+        + gradient_string
+        + "</svg>"
+    )
+    svg = SVG.fromstring(svg_string)._apply_gradient_translation()
+    el = svg.xpath_one("//svg:linearGradient | //svg:radialGradient")
+
+    for node in svg.svg_root.getiterator():
+        node.tag = etree.QName(node).localname
+    etree.cleanup_namespaces(svg.svg_root)
+
+    assert etree.tostring(el).decode("utf-8") == expected_result
